.TaxonomySelect_option__vY6v2 strong{font-weight:400}.TaxonomySelect_option__vY6v2 em{visibility:hidden}.TaxonomySelect_option__vY6v2:hover strong{font-weight:600}.TaxonomySelect_option__vY6v2:hover em{visibility:visible}.EditMinimalDataUseModal_advancedSettings___i9jS{border:1px solid var(--ant-color-border)}