(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[8012],{43376:function(e,t,s){(window.__NEXT_P=window.__NEXT_P||[]).push(["/add-systems/manual",function(){return s(33688)}])},33688:function(e,t,s){"use strict";s.r(t);var a=s(24246),r=s(77181),n=s(86677),i=s(27378),o=s(16134),u=s(7617),d=s(77213),l=s(77830),m=s(58754),c=s(41337),y=s(57865),h=s(72774),b=s(84450);let x=e=>{let{connector:t}=e;return(0,a.jsxs)(r.xuv,{display:"flex",mb:4,alignItems:"center","data-testid":"header",children:[(0,a.jsx)(c.Z,{data:null!=t?t:"ethyca",mr:2}),(0,a.jsxs)(r.X6q,{fontSize:"md",children:["Describe your ",t?t.human_readable:"new"," system"]})]})};t.default=()=>{let{connectorType:e}=(0,n.useRouter)().query,{tabData:t,tabIndex:s,onTabChange:c}=(0,b.Z)({isCreate:!0}),f=(0,i.useMemo)(()=>{if(e)return JSON.parse(Array.isArray(e)?e[0]:e)},[e]),_=(0,o.C)(y.gU);return(0,a.jsxs)(d.Z,{title:"Describe your system",children:[(0,a.jsx)(m.Z,{heading:"Add systems",breadcrumbItems:[{title:"Add systems",href:l.xo},{title:"New system"}]}),(0,a.jsx)(x,{connector:f}),_?(0,a.jsx)(h.Z,{}):null,(0,a.jsxs)(r.xuv,{w:{base:"100%",md:"75%"},children:[(0,a.jsx)(r.xvT,{fontSize:"sm",mb:8,children:"Systems are anything that might store or process data in your organization, from a web application, to a database or data warehouse. Describe your system below to register it to the map. You may optionally complete data entry for the system using the additional tabs to navigate the sections."}),(0,a.jsx)(u.Z,{data:t,"data-testid":"system-tabs",index:s,isLazy:!0,isManual:!0,onChange:c})]})]})}}},function(e){e.O(0,[8033,6060,2858,2866,9278,5277,4481,7980,355,8499,7453,4450,2888,9774,179],function(){return e(e.s=43376)}),_N_E=e.O()}]);