"""add modal_link_label to experience translation model

Revision ID: e4023342ebbb
Revises: 7799269e7078
Create Date: 2024-04-02 01:16:46.183952

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e4023342ebbb"
down_revision = "7799269e7078"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experiencetranslation",
        sa.Column("modal_link_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("modal_link_label", sa.String(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfighistory", "modal_link_label")
    op.drop_column("experiencetranslation", "modal_link_label")
    # ### end Alembic commands ###
