"""
emailconfig

Revision ID: c61fd9d4f73e
Revises: 7abe778b7082
Create Date: 2022-08-10 19:29:12.119401

"""

from alembic import op

# revision identifiers, used by Alembic.

revision = "c61fd9d4f73e"
down_revision = "7abe778b7082"
branch_labels = None
depends_on = None


def upgrade():
    # Removing due to the following error: type "emailservicetype" already exists
    pass


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_emailconfig_service_type"), table_name="emailconfig")
    op.drop_index(op.f("ix_emailconfig_name"), table_name="emailconfig")
    op.drop_index(op.f("ix_emailconfig_key"), table_name="emailconfig")
    op.drop_index(op.f("ix_emailconfig_id"), table_name="emailconfig")
    op.drop_table("emailconfig")
    # ### end Alembic commands ###
