(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2357],{5123:function(e,n,i){(window.__NEXT_P=window.__NEXT_P||[]).push(["/consent/privacy-experience/[id]",function(){return i(41884)}])},41884:function(e,n,i){"use strict";i.r(n);var r=i(24246),c=i(77181),t=i(86677),s=i(77213),d=i(3412),u=i(94315);n.default=()=>{let e=(0,t.useRouter)(),n="";e.query.id&&(n=Array.isArray(e.query.id)?e.query.id[0]:e.query.id);let{data:i,isLoading:a}=(0,u.Qv)(n),{data:o,isLoading:p}=(0,u.fc)(n);return a||p?(0,r.jsx)(s.Z,{title:"Privacy experience",children:(0,r.jsx)(c.M5Y,{children:(0,r.jsx)(c.$jN,{})})}):i?(0,r.jsx)(s.Z,{title:"Privacy experience ".concat(i.component),padded:!1,children:(0,r.jsx)(d.Z,{passedInExperience:i,passedInTranslations:o})}):(0,r.jsx)(s.Z,{title:"Privacy experience",children:(0,r.jsxs)(c.xvT,{children:["No privacy experience with id ",n," found."]})})}}},function(e){e.O(0,[2858,2866,4294,9014,9282,1376,3412,2888,9774,179],function(){return e(e.s=5123)}),_N_E=e.O()}]);