from fides.api.service.connectors.erasure_email_connector import (
    GenericErasureEmailConnector,
)


class AttentiveConnector(GenericErasureEmailConnector):
    """No custom logic here but we want to consider this as a separate connector"""
